"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const constants_1 = require("./constants");
const parseHeight = (height) => {
    if (typeof height === 'undefined' ||
        height === null ||
        Number.isNaN(height)) {
        throw new TypeError('Invalid block height');
    }
    return Math.abs(height);
};
exports.calculateMilestone = (height, blockRewardArgs) => {
    const parsedHeight = parseHeight(height);
    const distance = Math.floor(blockRewardArgs.distance);
    const location = Math.trunc((parsedHeight - blockRewardArgs.rewardOffset) / distance);
    const lastMile = blockRewardArgs.milestones[blockRewardArgs.milestones.length - 1];
    if (location > blockRewardArgs.milestones.length - 1) {
        return blockRewardArgs.milestones.lastIndexOf(lastMile);
    }
    return location;
};
exports.calculateReward = (height, blockRewardArgs) => {
    const parsedHeight = parseHeight(height);
    if (parsedHeight < blockRewardArgs.rewardOffset) {
        return BigInt(0);
    }
    return BigInt(blockRewardArgs.milestones[exports.calculateMilestone(parsedHeight, blockRewardArgs)]);
};
exports.calculateSupply = (height, blockRewardArgs) => {
    let parsedHeight = parseHeight(height);
    const distance = Math.floor(blockRewardArgs.distance);
    let supply = BigInt(blockRewardArgs.totalAmount);
    if (parsedHeight < blockRewardArgs.rewardOffset) {
        return supply;
    }
    const milestone = exports.calculateMilestone(parsedHeight, blockRewardArgs);
    const rewards = [];
    let amount = 0;
    let multiplier;
    parsedHeight -= blockRewardArgs.rewardOffset - 1;
    for (let i = 0; i < blockRewardArgs.milestones.length; i += 1) {
        if (milestone >= i) {
            multiplier = blockRewardArgs.milestones[i];
            if (parsedHeight < distance) {
                amount = parsedHeight % distance;
            }
            else {
                amount = distance;
                parsedHeight -= distance;
                if (parsedHeight > 0 && i === blockRewardArgs.milestones.length - 1) {
                    amount += parsedHeight;
                }
            }
            rewards.push([amount, multiplier]);
        }
        else {
            break;
        }
    }
    for (let i = 0; i < rewards.length; i += 1) {
        const reward = rewards[i];
        supply = supply + BigInt(reward[0]) * BigInt(reward[1]);
    }
    return supply;
};
exports.getTotalFees = (blockInstance) => blockInstance.transactions.reduce((prev, current) => {
    const minFee = current.minFee;
    return {
        totalFee: prev.totalFee + current.fee,
        totalMinFee: prev.totalMinFee + minFee,
    };
}, { totalFee: BigInt(0), totalMinFee: BigInt(0) });
exports.applyFeeAndRewards = async (blockInstance, stateStore) => {
    const generatorAddress = lisk_cryptography_1.getAddressFromPublicKey(blockInstance.generatorPublicKey);
    const generator = await stateStore.account.get(generatorAddress);
    generator.balance += blockInstance.reward;
    generator.producedBlocks += 1;
    if (!blockInstance.transactions.length) {
        stateStore.account.set(generatorAddress, generator);
        return;
    }
    const { totalFee, totalMinFee } = exports.getTotalFees(blockInstance);
    const givenFee = totalFee - totalMinFee;
    generator.balance += givenFee > 0 ? givenFee : BigInt(0);
    const totalFeeBurntStr = await stateStore.chain.get(constants_1.CHAIN_STATE_BURNT_FEE);
    let totalFeeBurnt = BigInt(totalFeeBurntStr || 0);
    totalFeeBurnt += givenFee > 0 ? totalMinFee : BigInt(0);
    stateStore.account.set(generatorAddress, generator);
    stateStore.chain.set(constants_1.CHAIN_STATE_BURNT_FEE, totalFeeBurnt.toString());
};
exports.undoFeeAndRewards = async (blockInstance, stateStore) => {
    const generatorAddress = lisk_cryptography_1.getAddressFromPublicKey(blockInstance.generatorPublicKey);
    const generator = await stateStore.account.get(generatorAddress);
    generator.balance -= blockInstance.reward;
    generator.producedBlocks -= 1;
    if (!blockInstance.transactions.length) {
        stateStore.account.set(generatorAddress, generator);
        return;
    }
    const { totalFee, totalMinFee } = exports.getTotalFees(blockInstance);
    generator.balance -= totalFee - totalMinFee;
    const totalFeeBurntStr = await stateStore.chain.get(constants_1.CHAIN_STATE_BURNT_FEE);
    let totalFeeBurnt = BigInt(totalFeeBurntStr || 0);
    totalFeeBurnt -= totalMinFee;
    stateStore.account.set(generatorAddress, generator);
    stateStore.chain.set(constants_1.CHAIN_STATE_BURNT_FEE, totalFeeBurnt.toString());
};
//# sourceMappingURL=block_reward.js.map