"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const account_1 = require("../account");
const cache_1 = require("./cache");
const storage_1 = require("./storage");
const transaction_interface_adapter_1 = require("./transaction_interface_adapter");
class DataAccess {
    constructor({ dbStorage, registeredTransactions, minBlockHeaderCache, maxBlockHeaderCache, }) {
        this._storage = new storage_1.Storage(dbStorage);
        this._blocksCache = new cache_1.BlockCache(minBlockHeaderCache, maxBlockHeaderCache);
        this._transactionAdapter = new transaction_interface_adapter_1.TransactionInterfaceAdapter(registeredTransactions);
    }
    addBlockHeader(blockHeader) {
        return this._blocksCache.add(blockHeader);
    }
    async removeBlockHeader(id) {
        var _a;
        const cachedItems = this._blocksCache.remove(id);
        if (!this._blocksCache.needsRefill) {
            return cachedItems;
        }
        const upperHeightToFetch = ((_a = this._blocksCache.items[0]) === null || _a === void 0 ? void 0 : _a.height) - 1 || 0;
        const lowerHeightToFetch = Math.max(upperHeightToFetch -
            (this._blocksCache.maxCachedItems - this._blocksCache.minCachedItems), 1);
        if (upperHeightToFetch - lowerHeightToFetch > 0) {
            const blockHeaders = await this.getBlocksByHeightBetween(lowerHeightToFetch, upperHeightToFetch);
            this._blocksCache.refill(blockHeaders.reverse());
        }
        return cachedItems;
    }
    resetBlockHeaderCache() {
        this._blocksCache.empty();
    }
    async getBlockHeaderByID(id) {
        const cachedBlock = this._blocksCache.getByID(id);
        if (cachedBlock) {
            return cachedBlock;
        }
        const blockJSON = await this._storage.getBlockHeaderByID(id);
        return this.deserializeBlockHeader(blockJSON);
    }
    async getBlockHeadersByIDs(arrayOfBlockIds) {
        const cachedBlocks = this._blocksCache.getByIDs(arrayOfBlockIds);
        if (cachedBlocks.length) {
            return cachedBlocks;
        }
        const blocks = await this._storage.getBlockHeadersByIDs(arrayOfBlockIds);
        return blocks.map(block => this.deserializeBlockHeader(block));
    }
    async getBlockHeaderByHeight(height) {
        const cachedBlock = this._blocksCache.getByHeight(height);
        if (cachedBlock) {
            return cachedBlock;
        }
        const block = await this._storage.getBlockByHeight(height);
        return block ? this.deserializeBlockHeader(block) : undefined;
    }
    async getBlockHeadersByHeightBetween(fromHeight, toHeight) {
        const cachedBlocks = this._blocksCache.getByHeightBetween(fromHeight, toHeight);
        if (cachedBlocks.length) {
            return cachedBlocks;
        }
        const blocks = await this._storage.getBlockHeadersByHeightBetween(fromHeight, toHeight);
        return blocks.map(block => this.deserializeBlockHeader(block));
    }
    async getBlockHeadersWithHeights(heightList) {
        const cachedBlocks = this._blocksCache.getByHeights(heightList);
        if (cachedBlocks.length) {
            return cachedBlocks;
        }
        const blocks = await this._storage.getBlockHeadersWithHeights(heightList);
        return blocks.map(block => this.deserializeBlockHeader(block));
    }
    async getLastBlockHeader() {
        const cachedBlock = this._blocksCache.last;
        if (cachedBlock) {
            return cachedBlock;
        }
        const block = await this._storage.getLastBlockHeader();
        return this.deserializeBlockHeader(block);
    }
    async getLastCommonBlockHeader(arrayOfBlockIds) {
        const blocks = this._blocksCache.getByIDs(arrayOfBlockIds);
        const cachedBlock = blocks[blocks.length - 1];
        if (cachedBlock) {
            return cachedBlock;
        }
        const block = await this._storage.getLastCommonBlockHeader(arrayOfBlockIds);
        return block ? this.deserializeBlockHeader(block) : undefined;
    }
    async getBlocksCount() {
        const blocksCount = await this._storage.getBlocksCount();
        return blocksCount;
    }
    async getBlockByID(id) {
        const blockJSON = await this._storage.getBlockByID(id);
        return blockJSON ? this.deserialize(blockJSON) : undefined;
    }
    async getBlocksByIDs(arrayOfBlockIds) {
        const blocks = await this._storage.getBlocksByIDs(arrayOfBlockIds);
        return blocks.map(block => this.deserialize(block));
    }
    async getBlockByHeight(height) {
        const block = await this._storage.getBlockByHeight(height);
        return block ? this.deserialize(block) : undefined;
    }
    async getBlocksByHeightBetween(fromHeight, toHeight) {
        const blocks = await this._storage.getBlocksByHeightBetween(fromHeight, toHeight);
        return blocks.map(block => this.deserialize(block));
    }
    async getBlocksWithLimitAndOffset(limit, offset = 0) {
        const toHeight = offset + limit;
        const toHeightLT = toHeight - 1;
        const blocks = await this.getBlocksByHeightBetween(offset, toHeightLT);
        const sortedBlocks = [...blocks].sort((a, b) => a.height - b.height);
        return sortedBlocks;
    }
    async getLastBlock() {
        const block = await this._storage.getLastBlock();
        return block && this.deserialize(block);
    }
    async deleteBlocksWithHeightGreaterThan(height) {
        await this._storage.deleteBlocksWithHeightGreaterThan(height);
    }
    async isBlockPersisted(blockId) {
        const isPersisted = await this._storage.isBlockPersisted(blockId);
        return isPersisted;
    }
    async getTempBlocks() {
        const blocks = await this._storage.getTempBlocks();
        return blocks;
    }
    async isTempBlockEmpty() {
        const isEmpty = await this._storage.isTempBlockEmpty();
        return isEmpty;
    }
    clearTempBlocks() {
        this._storage.clearTempBlocks();
    }
    async getChainState(key) {
        return this._storage.getChainState(key);
    }
    async getConsensusState(key) {
        return this._storage.getConsensusState(key);
    }
    async getAccountsByPublicKey(arrayOfPublicKeys) {
        const accounts = await this._storage.getAccountsByPublicKey(arrayOfPublicKeys);
        return accounts.map(account => new account_1.Account(account));
    }
    async getAccountByAddress(address) {
        const account = await this._storage.getAccountByAddress(address);
        return new account_1.Account(account);
    }
    async getAccountsByAddress(arrayOfAddresses) {
        const accounts = await this._storage.getAccountsByAddress(arrayOfAddresses);
        return accounts.map(account => new account_1.Account(account));
    }
    async getDelegates() {
        const accounts = await this._storage.getDelegates();
        return accounts.map(account => new account_1.Account(account));
    }
    async resetAccountMemTables() {
        await this._storage.resetAccountMemTables();
    }
    async getTransactionsByIDs(arrayOfTransactionIds) {
        const transactions = await this._storage.getTransactionsByIDs(arrayOfTransactionIds);
        return transactions.map(transaction => this.deserializeTransaction(transaction));
    }
    async isTransactionPersisted(transactionId) {
        const isPersisted = await this._storage.isTransactionPersisted(transactionId);
        return isPersisted;
    }
    serialize(blockInstance) {
        const { transactions, ...blockHeader } = blockInstance;
        const blockHeaderJSON = this.serializeBlockHeader(blockHeader);
        const transactionsJSON = transactions.map(tx => ({
            ...tx.toJSON(),
            blockId: blockInstance.id,
        }));
        return {
            ...blockHeaderJSON,
            transactions: transactionsJSON,
        };
    }
    deserialize(blockJSON) {
        var _a, _b;
        const transactions = (_b = (_a = blockJSON.transactions) === null || _a === void 0 ? void 0 : _a.map(transaction => this._transactionAdapter.fromJSON(transaction))) !== null && _b !== void 0 ? _b : [];
        return {
            ...blockJSON,
            totalAmount: BigInt(blockJSON.totalAmount || 0),
            totalFee: BigInt(blockJSON.totalFee || 0),
            reward: BigInt(blockJSON.reward || 0),
            transactions,
        };
    }
    serializeBlockHeader(blockHeader) {
        return {
            ...blockHeader,
            totalAmount: blockHeader.totalAmount.toString(),
            totalFee: blockHeader.totalFee.toString(),
            reward: blockHeader.reward.toString(),
        };
    }
    deserializeBlockHeader(blockHeader) {
        return {
            ...blockHeader,
            totalAmount: BigInt(blockHeader.totalAmount || 0),
            totalFee: BigInt(blockHeader.totalFee || 0),
            reward: BigInt(blockHeader.reward || 0),
        };
    }
    deserializeTransaction(transactionJSON) {
        return this._transactionAdapter.fromJSON(transactionJSON);
    }
}
exports.DataAccess = DataAccess;
//# sourceMappingURL=data_access.js.map