"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
exports.validateTransactions = () => (transactions) => transactions.map(transaction => transaction.validate());
exports.applyGenesisTransactions = () => async (transactions, stateStore) => {
    for (const transaction of transactions) {
        await transaction.prepare(stateStore);
    }
    const transactionsResponses = [];
    for (const transaction of transactions) {
        const transactionResponse = await transaction.apply(stateStore);
        transactionResponse.status =
            lisk_transactions_1.Status.OK;
        transactionsResponses.push(transactionResponse);
    }
    return transactionsResponses;
};
exports.applyTransactions = () => async (transactions, stateStore) => {
    for (const transaction of transactions) {
        await transaction.prepare(stateStore);
    }
    const transactionsResponses = [];
    for (const transaction of transactions) {
        stateStore.account.createSnapshot();
        const transactionResponse = await transaction.apply(stateStore);
        if (transactionResponse.status !== lisk_transactions_1.Status.OK) {
            stateStore.account.restoreSnapshot();
        }
        transactionsResponses.push(transactionResponse);
    }
    return [...transactionsResponses];
};
exports.checkPersistedTransactions = (dataAccess) => async (transactions) => {
    if (!transactions.length) {
        return [];
    }
    const confirmedTransactions = await dataAccess.getTransactionsByIDs(transactions.map(transaction => transaction.id));
    const persistedTransactionIds = confirmedTransactions.map((transaction) => transaction.id);
    const persistedTransactions = transactions.filter(transaction => persistedTransactionIds.includes(transaction.id));
    const nonPersistedTransactions = transactions.filter(transaction => !persistedTransactionIds.includes(transaction.id));
    const transactionsResponses = [
        ...nonPersistedTransactions.map(transaction => ({
            id: transaction.id,
            status: lisk_transactions_1.Status.OK,
            errors: [],
        })),
        ...persistedTransactions.map(transaction => ({
            id: transaction.id,
            status: lisk_transactions_1.Status.FAIL,
            errors: [
                new lisk_transactions_1.TransactionError(`Transaction is already confirmed: ${transaction.id}`, transaction.id, '.id'),
            ],
        })),
    ];
    return transactionsResponses;
};
exports.checkAllowedTransactions = (contexter) => (transactions) => transactions.map(transaction => {
    const context = typeof contexter === 'function' ? contexter() : contexter;
    const allowed = !transaction.matcher ||
        transaction.matcher(context);
    return {
        id: transaction.id,
        status: allowed ? lisk_transactions_1.Status.OK : lisk_transactions_1.Status.FAIL,
        errors: allowed
            ? []
            : [
                new lisk_transactions_1.TransactionError(`Transaction type ${transaction.type} is currently not allowed.`, transaction.id),
            ],
    };
});
exports.undoTransactions = () => async (transactions, stateStore) => {
    for (const transaction of transactions) {
        await transaction.prepare(stateStore);
    }
    const transactionsResponses = [];
    for (const transaction of transactions) {
        const transactionResponse = await transaction.undo(stateStore);
        transactionsResponses.push(transactionResponse);
    }
    return transactionsResponses;
};
//# sourceMappingURL=transactions_handlers.js.map