"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const Debug = require("debug");
const events_1 = require("events");
const constants_1 = require("./constants");
const delegates_info_1 = require("./delegates_info");
const delegates_list_1 = require("./delegates_list");
const rounds_1 = require("./rounds");
const debug = Debug('lisk:dpos');
class Dpos {
    constructor({ chain, activeDelegates = constants_1.DEFAULT_ACTIVE_DELEGATE, standbyDelegates = constants_1.DEFAULT_STANDBY_DELEGATE, standbyThreshold = constants_1.DEFAULT_STANDBY_THRESHOLD, delegateListRoundOffset = constants_1.DEFAULT_ROUND_OFFSET, voteWeightCapRate = constants_1.DEFAULT_VOTE_WEIGHT_CAP_RATE, }) {
        this.events = new events_1.EventEmitter();
        this.delegateListRoundOffset = delegateListRoundOffset;
        this._activeDelegates = activeDelegates;
        this._delegatesPerRound = activeDelegates + standbyDelegates;
        this.delegateActiveRoundLimit = 3;
        this.chain = chain;
        this.rounds = new rounds_1.Rounds({
            blocksPerRound: this._delegatesPerRound,
        });
        this.delegatesList = new delegates_list_1.DelegatesList({
            rounds: this.rounds,
            activeDelegates,
            standbyDelegates,
            standbyThreshold,
            voteWeightCapRate,
            chain: this.chain,
        });
        this.delegatesInfo = new delegates_info_1.DelegatesInfo({
            chain: this.chain,
            rounds: this.rounds,
            activeDelegates,
            standbyDelegates,
            events: this.events,
            delegatesList: this.delegatesList,
        });
    }
    get delegatesPerRound() {
        return this._delegatesPerRound;
    }
    async getForgerAddressesForRound(round) {
        return this.delegatesList.getDelegateList(round);
    }
    async onBlockFinalized(stateStore, finalizedHeight) {
        const finalizedBlockRound = this.rounds.calcRound(finalizedHeight);
        const disposableDelegateList = finalizedBlockRound -
            this.delegateListRoundOffset -
            this.delegateActiveRoundLimit;
        await delegates_list_1.deleteForgersListUntilRound(disposableDelegateList, stateStore);
        await delegates_list_1.deleteVoteWeightsUntilRound(disposableDelegateList, stateStore);
    }
    async getMinActiveHeight(height, address, stateStore, delegateActiveRoundLimit) {
        const forgersList = await delegates_list_1.getForgersList(stateStore);
        if (!forgersList.length) {
            throw new Error('No delegate list found in the database.');
        }
        forgersList.sort((a, b) => b.round - a.round);
        const currentRound = this.rounds.calcRound(height);
        const previousForgersList = forgersList.filter(fl => fl.round < currentRound);
        const activeRounds = this._findEarliestActiveListRound(address, previousForgersList, delegateActiveRoundLimit);
        return this.rounds.calcRoundStartHeight(activeRounds);
    }
    async isActiveDelegate(address, height) {
        const relevantRound = this.rounds.calcRound(height);
        const voteWeightsStr = await this.chain.dataAccess.getConsensusState(constants_1.CONSENSUS_STATE_VOTE_WEIGHTS_KEY);
        const voteWeights = voteWeightsStr
            ? JSON.parse(voteWeightsStr)
            : [];
        const voteWeight = voteWeights.find(roundRecord => roundRecord.round === relevantRound);
        if (!voteWeight) {
            throw new Error(`Vote weight not found for round ${relevantRound} for the given height ${height}`);
        }
        const activeDelegateVoteWeights = voteWeight.delegates.slice(0, this._activeDelegates);
        return (activeDelegateVoteWeights.findIndex(vw => vw.address === address) > -1);
    }
    async isStandbyDelegate(address, height, stateStore) {
        const forgersList = await delegates_list_1.getForgersList(stateStore);
        const relevantRound = this.rounds.calcRound(height);
        const foundForgerList = forgersList.find(roundRecord => roundRecord.round === relevantRound);
        if (!foundForgerList) {
            throw new Error(`Forger list not found for round ${relevantRound} for the given height ${height}`);
        }
        const isStandby = foundForgerList.standby.find(standByDelegate => standByDelegate === address);
        return isStandby ? true : false;
    }
    _findEarliestActiveListRound(address, previousLists, delegateActiveRoundLimit = this.delegateActiveRoundLimit) {
        if (!previousLists.length) {
            return 0;
        }
        const lists = previousLists.slice(0, delegateActiveRoundLimit);
        for (let i = 0; i < lists.length; i += 1) {
            const { round, delegates } = lists[i];
            if (delegates.indexOf(address) === -1) {
                return round + 1;
            }
        }
        return lists[lists.length - 1].round;
    }
    async verifyBlockForger(block) {
        return this.delegatesList.verifyBlockForger(block);
    }
    async apply(block, stateStore, { delegateListRoundOffset } = {
        delegateListRoundOffset: this.delegateListRoundOffset,
    }) {
        return this.delegatesInfo.apply(block, stateStore, {
            delegateListRoundOffset,
        });
    }
    async undo(block, stateStore, { delegateListRoundOffset } = {
        delegateListRoundOffset: this.delegateListRoundOffset,
    }) {
        return this.delegatesInfo.undo(block, stateStore, {
            delegateListRoundOffset,
        });
    }
    async isDPoSProtocolCompliant(blockHeader, store) {
        const { lastBlockHeaders } = store.consensus;
        const round = this.rounds.calcRound(blockHeader.height);
        const startOfLastRound = this.rounds.calcRoundStartHeight(round - 1);
        const delegateForgedBlocks = lastBlockHeaders.filter(block => block.generatorPublicKey === blockHeader.generatorPublicKey &&
            block.height >= startOfLastRound);
        if (!delegateForgedBlocks.length) {
            debug('Delegate did not forge any block in current or last round', {
                generatorPublicKey: blockHeader.generatorPublicKey,
                height: blockHeader.height,
            });
            return true;
        }
        const { seedReveal: previousBlockSeedReveal } = delegateForgedBlocks[0];
        const { seedReveal: newBlockSeedReveal } = blockHeader;
        const SEED_REVEAL_BYTE_SIZE = 16;
        const newBlockSeedRevealBuffer = lisk_cryptography_1.hash(lisk_cryptography_1.hexToBuffer(newBlockSeedReveal)).slice(0, SEED_REVEAL_BYTE_SIZE);
        if (lisk_cryptography_1.hexToBuffer(previousBlockSeedReveal).equals(newBlockSeedRevealBuffer)) {
            return true;
        }
        debug('New block SeedReveal is not the preimage of last block', {
            newBlockSeedReveal: newBlockSeedRevealBuffer.toString('hex'),
            previousBlockSeedReveal,
            delegate: blockHeader.generatorPublicKey,
            height: blockHeader.height,
        });
        return false;
    }
}
exports.Dpos = Dpos;
//# sourceMappingURL=dpos.js.map