"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_validator_1 = require("@liskhq/lisk-validator");
const _12_multisignature_transaction_1 = require("./12_multisignature_transaction");
const constants_1 = require("./constants");
const utils_1 = require("./utils");
const validateInputs = ({ mandatoryPublicKeys, optionalPublicKeys, numberOfSignatures, networkIdentifier, fee, nonce, }) => {
    if (!lisk_validator_1.isValidNonce(nonce)) {
        throw new Error('Nonce must be a valid number in string format.');
    }
    if (!lisk_validator_1.isValidFee(fee)) {
        throw new Error('Fee must be a valid number in string format.');
    }
    if (!lisk_validator_1.isValidInteger(numberOfSignatures) ||
        numberOfSignatures < constants_1.MIN_NUMBER_OF_SIGNATURES ||
        numberOfSignatures > constants_1.MAX_NUMBER_OF_SIGNATURES) {
        throw new Error(`Please provide a valid numberOfSignatures value. Expected integer between ${constants_1.MIN_NUMBER_OF_SIGNATURES} and ${constants_1.MAX_NUMBER_OF_SIGNATURES}.`);
    }
    if (mandatoryPublicKeys.length > numberOfSignatures) {
        throw new Error('The numberOfSignatures should be more than or equal to the number of mandatory keys.');
    }
    if (numberOfSignatures >
        mandatoryPublicKeys.length + optionalPublicKeys.length) {
        throw new Error(`Please provide a valid numberOfSignatures. numberOfSignatures (${numberOfSignatures}) is bigger than the count of optional (${optionalPublicKeys.length}) and mandatory (${mandatoryPublicKeys.length}) keys.`);
    }
    if (mandatoryPublicKeys.length + optionalPublicKeys.length >
        constants_1.MAX_NUMBER_OF_KEYS ||
        mandatoryPublicKeys.length + optionalPublicKeys.length < constants_1.MIN_NUMBER_OF_KEYS) {
        throw new Error(`Please provide a valid number of mandatory and optional keys. Expected integer between ${constants_1.MIN_NUMBER_OF_SIGNATURES} and ${constants_1.MAX_NUMBER_OF_SIGNATURES}.`);
    }
    const repeatedKeys = utils_1.findRepeatedKeys(optionalPublicKeys, mandatoryPublicKeys);
    if (repeatedKeys.length > 0) {
        throw new Error(`There are repeated values in optional and mandatory keys: '${repeatedKeys.join(', ')}'`);
    }
    const uniqueKeys = Array.from(new Set([...mandatoryPublicKeys, ...optionalPublicKeys]));
    if (uniqueKeys.length !==
        mandatoryPublicKeys.length + optionalPublicKeys.length) {
        throw new Error('There are repeated public keys. Mandatory and Optional Public Keys need too be unique.');
    }
    lisk_validator_1.validateNetworkIdentifier(networkIdentifier);
};
exports.registerMultisignature = (inputs) => {
    const { senderPassphrase, passphrases, mandatoryKeys, optionalKeys, numberOfSignatures, networkIdentifier, fee, nonce, } = inputs;
    validateInputs({
        mandatoryPublicKeys: mandatoryKeys,
        optionalPublicKeys: optionalKeys,
        numberOfSignatures,
        networkIdentifier,
        fee,
        nonce,
    });
    const transaction = {
        ...utils_1.createBaseTransaction(inputs),
        type: 12,
        asset: {
            mandatoryKeys,
            optionalKeys,
            numberOfSignatures,
        },
    };
    const multisignatureTransaction = new _12_multisignature_transaction_1.MultisignatureTransaction(transaction);
    if (!passphrases || !senderPassphrase) {
        return multisignatureTransaction.toJSON();
    }
    multisignatureTransaction.sign(networkIdentifier, senderPassphrase, passphrases, {
        mandatoryKeys,
        optionalKeys,
        numberOfSignatures,
    });
    return multisignatureTransaction.toJSON();
};
//# sourceMappingURL=register_multisignature_account.js.map