"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../../base"));
const error_1 = require("../../../utils/error");
const flags_1 = require("../../../utils/flags");
const network_identifier_1 = require("../../../utils/network_identifier");
const reader_1 = require("../../../utils/reader");
const dataFlag = {
    char: 'd',
    description: `Optional UTF8 encoded data (maximum of 64 bytes) to include in the transaction asset.
	Examples:
	- --data=customInformation
`,
};
const processInputs = (nonce, fee, networkIdentifier, amount, address, data, passphrase) => lisk_transactions_1.transfer({
    nonce,
    fee,
    networkIdentifier,
    recipientId: address,
    amount,
    data,
    passphrase,
});
class TransferCommand extends base_1.default {
    async run() {
        const { args, flags: { networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'no-signature': noSignature, data: dataString, }, } = this.parse(TransferCommand);
        const { nonce, fee, amount, address } = args;
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        if (!lisk_validator_1.isValidNonce(nonce)) {
            throw new error_1.ValidationError('Enter a valid nonce in number string format.');
        }
        if (Number.isNaN(Number(fee))) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const normalizedFee = lisk_transactions_1.utils.convertLSKToBeddows(fee);
        if (!lisk_validator_1.isValidFee(normalizedFee)) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        lisk_validator_1.validateAddress(address);
        const normalizedAmount = lisk_transactions_1.utils.convertLSKToBeddows(amount);
        if (noSignature) {
            const noSignatureResult = processInputs(nonce, normalizedFee, networkIdentifier, normalizedAmount, address, dataString);
            this.print(noSignatureResult);
            return;
        }
        const passphrase = passphraseSource !== null && passphraseSource !== void 0 ? passphraseSource : (await reader_1.getPassphraseFromPrompt('passphrase', true));
        const result = processInputs(nonce, normalizedFee, networkIdentifier, normalizedAmount, address, dataString, passphrase);
        this.print(result);
    }
}
exports.default = TransferCommand;
TransferCommand.args = [
    {
        name: 'nonce',
        required: true,
        description: 'Nonce of the transaction.',
    },
    {
        name: 'fee',
        required: true,
        description: 'Transaction fee in LSK.',
    },
    {
        name: 'amount',
        required: true,
        description: 'Amount of LSK to send.',
    },
    {
        name: 'address',
        required: true,
        description: 'Address of the recipient.',
    },
];
TransferCommand.description = `
	Creates a transaction which will transfer the specified amount to an address if broadcast to the network.
		`;
TransferCommand.examples = [
    'transaction:create:transfer 1 100 100 13356260975429434553L',
];
TransferCommand.flags = {
    ...base_1.default.flags,
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    passphrase: command_1.flags.string(flags_1.flags.passphrase),
    'no-signature': command_1.flags.boolean(flags_1.flags.noSignature),
    data: command_1.flags.string(dataFlag),
};
//# sourceMappingURL=transfer.js.map