"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const liskPassphrase = tslib_1.__importStar(require("@liskhq/lisk-passphrase"));
const fs_1 = tslib_1.__importDefault(require("fs"));
const inquirer_1 = tslib_1.__importDefault(require("inquirer"));
const readline_1 = tslib_1.__importDefault(require("readline"));
const error_1 = require("./error");
const capitalise = (text) => `${text.charAt(0).toUpperCase()}${text.slice(1)}`;
const getPassphraseVerificationFailError = (displayName) => `${capitalise(displayName)} was not successfully repeated.`;
const splitSource = (source) => {
    const delimiter = ':';
    const sourceParts = source.split(delimiter);
    return {
        sourceType: sourceParts[0],
        sourceIdentifier: sourceParts.slice(1).join(delimiter),
    };
};
exports.getPassphraseFromPrompt = async (displayName = 'passphrase', shouldConfirm = false) => {
    const questions = [
        {
            type: 'password',
            name: 'passphrase',
            message: `Please enter ${displayName}: `,
        },
    ];
    if (shouldConfirm) {
        questions.push({
            type: 'password',
            name: 'passphraseRepeat',
            message: `Please re-enter ${displayName}: `,
        });
    }
    const { passphrase, passphraseRepeat } = (await inquirer_1.default.prompt(questions));
    if (!passphrase || (shouldConfirm && passphrase !== passphraseRepeat)) {
        throw new error_1.ValidationError(getPassphraseVerificationFailError(displayName));
    }
    const passphraseErrors = [passphrase]
        .filter(Boolean)
        .map(pass => liskPassphrase.validation
        .getPassphraseValidationErrors(pass)
        .filter((error) => error.message));
    passphraseErrors.forEach(errors => {
        if (errors.length > 0) {
            const passphraseWarning = errors
                .filter((error) => error.code !== 'INVALID_MNEMONIC')
                .reduce((accumulator, error) => accumulator.concat(`${error.message.replace(' Please check the passphrase.', '')} `), 'Warning: ');
            console.warn(passphraseWarning);
        }
    });
    return passphrase;
};
const getFileDoesNotExistError = (path) => `File at ${path} does not exist.`;
const getFileUnreadableError = (path) => `File at ${path} could not be read.`;
const getDataFromFile = (path) => fs_1.default.readFileSync(path, 'utf8');
const ERROR_DATA_MISSING = 'No data was provided.';
const ERROR_DATA_SOURCE = 'Unknown data source type.';
exports.isFileSource = (source) => {
    if (!source) {
        return false;
    }
    const delimiter = ':';
    const sourceParts = source.split(delimiter);
    if (sourceParts.length === 2 && sourceParts[0] === 'file') {
        return true;
    }
    return false;
};
exports.readFileSource = async (source) => {
    if (!source) {
        throw new error_1.ValidationError(ERROR_DATA_MISSING);
    }
    const { sourceType, sourceIdentifier: path } = splitSource(source);
    if (sourceType !== 'file') {
        throw new error_1.ValidationError(ERROR_DATA_SOURCE);
    }
    try {
        return getDataFromFile(path);
    }
    catch (error) {
        const { message } = error;
        if (message.match(/ENOENT/)) {
            throw new error_1.FileSystemError(getFileDoesNotExistError(path));
        }
        if (message.match(/EACCES/)) {
            throw new error_1.FileSystemError(getFileUnreadableError(path));
        }
        throw error;
    }
};
const DEFAULT_TIMEOUT = 100;
exports.readStdIn = async () => {
    const readFromStd = new Promise((resolve, reject) => {
        const lines = [];
        const rl = readline_1.default.createInterface({ input: process.stdin });
        setTimeout(() => {
            reject(new Error(`Timed out after ${DEFAULT_TIMEOUT} ms`));
        }, DEFAULT_TIMEOUT);
        const handleClose = () => {
            resolve(lines);
        };
        return rl.on('line', line => lines.push(line)).on('close', handleClose);
    });
    return readFromStd;
};
//# sourceMappingURL=reader.js.map