"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const base_transaction_1 = require("./base_transaction");
const constants_1 = require("./constants");
const errors_1 = require("./errors");
const utils_1 = require("./utils");
const voteAssetFormatSchema = {
    type: 'object',
    required: ['votes'],
    properties: {
        votes: {
            type: 'array',
            minItems: 1,
            maxItems: 20,
            items: {
                type: 'object',
                required: ['delegateAddress', 'amount'],
                properties: {
                    delegateAddress: {
                        type: 'string',
                        format: 'address',
                    },
                    amount: {
                        type: 'string',
                        format: 'int64',
                    },
                },
            },
        },
    },
};
const SIZE_INT64 = 8;
const SIZE_UINT64 = SIZE_INT64;
const TEN_UNIT = BigInt(10) * BigInt(10) ** BigInt(8);
const MAX_VOTE = 10;
const MAX_UNLOCKING = 20;
class VoteTransaction extends base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        if (tx.asset) {
            const rawAsset = tx.asset;
            this.asset = {
                votes: rawAsset.votes.map(vote => {
                    const amount = lisk_validator_1.isNumberString(vote.amount)
                        ? BigInt(vote.amount)
                        : BigInt(0);
                    return {
                        delegateAddress: vote.delegateAddress,
                        amount,
                    };
                }),
            };
        }
        else {
            this.asset = { votes: [] };
        }
    }
    assetToJSON() {
        return {
            votes: this.asset.votes.map(vote => ({
                delegateAddress: vote.delegateAddress,
                amount: vote.amount.toString(),
            })),
        };
    }
    assetToBytes() {
        const bufferArray = [];
        for (const vote of this.asset.votes) {
            const addressBuffer = lisk_cryptography_1.intToBuffer(vote.delegateAddress.slice(0, -1), SIZE_UINT64);
            bufferArray.push(addressBuffer);
            const amountBuffer = lisk_cryptography_1.intToBuffer(vote.amount.toString(), SIZE_INT64, 'big', true);
            bufferArray.push(amountBuffer);
        }
        return Buffer.concat(bufferArray);
    }
    async prepare(store) {
        const addressArray = this.asset.votes.map(vote => ({
            address: vote.delegateAddress,
        }));
        const filterArray = [
            {
                address: this.senderId,
            },
            ...addressArray,
        ];
        await store.account.cache(filterArray);
    }
    validateAsset() {
        const asset = this.assetToJSON();
        const schemaErrors = lisk_validator_1.validator.validate(voteAssetFormatSchema, asset);
        const errors = errors_1.convertToAssetError(this.id, schemaErrors);
        let upvoteCount = 0;
        let downvoteCount = 0;
        const addressSet = new Set();
        for (const vote of this.asset.votes) {
            addressSet.add(vote.delegateAddress);
            if (vote.amount > BigInt(0)) {
                upvoteCount += 1;
            }
            else if (vote.amount < BigInt(0)) {
                downvoteCount += 1;
            }
            else {
                errors.push(new errors_1.TransactionError('Amount cannot be 0', this.id, '.asset.votes.amount'));
            }
            if (vote.amount % TEN_UNIT !== BigInt(0)) {
                errors.push(new errors_1.TransactionError('Amount should be multiple of 10 * 10^8', this.id, '.asset.votes.amount', vote.amount.toString()));
            }
        }
        if (upvoteCount > MAX_VOTE) {
            errors.push(new errors_1.TransactionError('Upvote can only be casted upto 10', this.id, '.asset.votes', upvoteCount, '10'));
        }
        if (downvoteCount > MAX_VOTE) {
            errors.push(new errors_1.TransactionError('Downvote can only be casted upto 10', this.id, '.asset.votes', upvoteCount, '10'));
        }
        if (addressSet.size !== this.asset.votes.length) {
            errors.push(new errors_1.TransactionError('Delegate address must be unique', this.id, '.asset.votes.delegateAddress'));
        }
        return errors;
    }
    async applyAsset(store) {
        const assetCopy = [...this.asset.votes];
        assetCopy.sort((a, b) => {
            const diff = a.amount - b.amount;
            if (diff > BigInt(0)) {
                return 1;
            }
            if (diff < BigInt(0)) {
                return -1;
            }
            return 0;
        });
        const errors = [];
        for (const vote of assetCopy) {
            const sender = await store.account.get(this.senderId);
            const votedDelegate = await store.account.getOrDefault(vote.delegateAddress);
            if (!votedDelegate.username) {
                errors.push(new errors_1.TransactionError('Voted delegate is not registered', this.id, '.asset.votes.delegateAddress'));
                continue;
            }
            if (vote.amount < BigInt(0)) {
                const originalUpvoteIndex = sender.votes.findIndex(senderVote => senderVote.delegateAddress === vote.delegateAddress);
                if (originalUpvoteIndex < 0) {
                    errors.push(new errors_1.TransactionError('Cannot cast downvote to delegate who is not upvoted', this.id, '.asset.votes.delegateAddress'));
                    continue;
                }
                sender.votes[originalUpvoteIndex].amount += vote.amount;
                if (sender.votes[originalUpvoteIndex].amount < BigInt(0)) {
                    errors.push(new errors_1.TransactionError('Cannot downvote more than upvoted', this.id, '.asset.votes.amount'));
                }
                if (sender.votes[originalUpvoteIndex].amount === BigInt(0)) {
                    sender.votes = sender.votes.filter(senderVote => senderVote.delegateAddress !== vote.delegateAddress);
                }
                sender.unlocking.push({
                    delegateAddress: vote.delegateAddress,
                    amount: BigInt(-1) * vote.amount,
                    unvoteHeight: store.chain.lastBlockHeader.height + 1,
                });
                utils_1.sortUnlocking(sender.unlocking);
                if (sender.unlocking.length > MAX_UNLOCKING) {
                    errors.push(new errors_1.TransactionError(`Cannot downvote which exceeds account.unlocking to have more than ${MAX_UNLOCKING}`, this.id, '.asset.votes'));
                }
            }
            else {
                const originalUpvoteIndex = sender.votes.findIndex(senderVote => senderVote.delegateAddress === vote.delegateAddress);
                const index = originalUpvoteIndex > -1 ? originalUpvoteIndex : sender.votes.length;
                const upvote = originalUpvoteIndex > -1
                    ? sender.votes[originalUpvoteIndex]
                    : {
                        delegateAddress: vote.delegateAddress,
                        amount: BigInt(0),
                    };
                upvote.amount += vote.amount;
                if (upvote.amount > BigInt(constants_1.MAX_INT64)) {
                    errors.push(new errors_1.TransactionError('Cannot upvote which exceeds int64', this.id, '.asset.votes.amount'));
                }
                sender.balance -= vote.amount;
                sender.votes[index] = upvote;
                sender.votes.sort((a, b) => a.delegateAddress.localeCompare(b.delegateAddress, 'en'));
                if (sender.votes.length > MAX_VOTE) {
                    errors.push(new errors_1.TransactionError(`Account can only vote upto ${MAX_VOTE}`, this.id, '.asset.votes.amount'));
                }
            }
            store.account.set(sender.address, sender);
            const delegate = await store.account.get(vote.delegateAddress);
            delegate.totalVotesReceived += vote.amount;
            store.account.set(delegate.address, delegate);
        }
        return errors;
    }
    async undoAsset(store) {
        const assetCopy = [...this.asset.votes];
        assetCopy.sort((a, b) => {
            const diff = b.amount - a.amount;
            if (diff > BigInt(0)) {
                return 1;
            }
            if (diff < BigInt(0)) {
                return -1;
            }
            return 0;
        });
        for (const vote of assetCopy) {
            const sender = await store.account.get(this.senderId);
            if (vote.amount < BigInt(0)) {
                const originalUpvoteIndex = sender.votes.findIndex(senderVote => senderVote.delegateAddress === vote.delegateAddress);
                const index = originalUpvoteIndex > -1 ? originalUpvoteIndex : sender.votes.length;
                const upvote = originalUpvoteIndex > -1
                    ? sender.votes[originalUpvoteIndex]
                    : {
                        delegateAddress: vote.delegateAddress,
                        amount: BigInt(0),
                    };
                upvote.amount += vote.amount * BigInt(-1);
                sender.votes[index] = upvote;
                const unlockingIndex = sender.unlocking.findIndex(unlock => unlock.delegateAddress === vote.delegateAddress &&
                    unlock.amount === vote.amount * BigInt(-1) &&
                    unlock.unvoteHeight === store.chain.lastBlockHeader.height + 1);
                if (unlockingIndex < 0) {
                    throw new Error('Invalid data. unlocking object should exist while undo');
                }
                sender.unlocking.splice(unlockingIndex, 1);
                sender.votes.sort((a, b) => a.delegateAddress.localeCompare(b.delegateAddress, 'en'));
                utils_1.sortUnlocking(sender.unlocking);
            }
            else {
                const originalUpvoteIndex = sender.votes.findIndex(senderVote => senderVote.delegateAddress === vote.delegateAddress);
                if (originalUpvoteIndex < 0) {
                    throw new Error('Invalid data. Upvote should exist while undo');
                }
                sender.votes[originalUpvoteIndex].amount -= vote.amount;
                if (sender.votes[originalUpvoteIndex].amount === BigInt(0)) {
                    sender.votes.splice(originalUpvoteIndex, 1);
                }
                sender.balance += vote.amount;
                sender.votes.sort((a, b) => a.delegateAddress.localeCompare(b.delegateAddress, 'en'));
            }
            store.account.set(sender.address, sender);
            const delegate = await store.account.get(vote.delegateAddress);
            delegate.totalVotesReceived += vote.amount * BigInt(-1);
            store.account.set(delegate.address, delegate);
        }
        return [];
    }
}
exports.VoteTransaction = VoteTransaction;
VoteTransaction.TYPE = 13;
//# sourceMappingURL=13_vote_transaction.js.map