"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const base_transaction_1 = require("./base_transaction");
const errors_1 = require("./errors");
const utils_1 = require("./utils");
const unlockAssetFormatSchema = {
    type: 'object',
    required: ['unlockingObjects'],
    properties: {
        unlockingObjects: {
            type: 'array',
            minItems: 1,
            maxItems: 20,
            items: {
                type: 'object',
                required: ['delegateAddress', 'amount', 'unvoteHeight'],
                properties: {
                    delegateAddress: {
                        type: 'string',
                        format: 'address',
                    },
                    amount: {
                        type: 'string',
                        format: 'int64',
                    },
                    unvoteHeight: {
                        type: 'integer',
                        minimum: 0,
                    },
                },
            },
        },
    },
};
const SIZE_UINT32 = 4;
const SIZE_INT64 = 8;
const SIZE_UINT64 = SIZE_INT64;
const AMOUNT_MULTIPLIER_FOR_VOTES = BigInt(10) * BigInt(10) ** BigInt(8);
const WAIT_TIME_VOTE = 2000;
const WAIT_TIME_SELF_VOTE = 260000;
const getWaitingPeriod = (sender, delegateAccount, lastBlockHeight, unlockObject) => {
    const currentHeight = lastBlockHeight + 1;
    const waitTime = sender.address === delegateAccount.address
        ? WAIT_TIME_SELF_VOTE
        : WAIT_TIME_VOTE;
    return waitTime - (currentHeight - unlockObject.unvoteHeight);
};
class UnlockTransaction extends base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        if (tx.asset) {
            const rawAsset = tx.asset;
            this.asset = {
                unlockingObjects: rawAsset.unlockingObjects.map(unlock => {
                    const amount = lisk_validator_1.isNumberString(unlock.amount)
                        ? BigInt(unlock.amount)
                        : BigInt(0);
                    return {
                        delegateAddress: unlock.delegateAddress,
                        amount,
                        unvoteHeight: unlock.unvoteHeight,
                    };
                }),
            };
        }
        else {
            this.asset = { unlockingObjects: [] };
        }
    }
    assetToJSON() {
        return {
            unlockingObjects: this.asset.unlockingObjects.map(unlock => ({
                delegateAddress: unlock.delegateAddress,
                amount: unlock.amount.toString(),
                unvoteHeight: unlock.unvoteHeight,
            })),
        };
    }
    assetToBytes() {
        const bufferArray = [];
        for (const unlock of this.asset.unlockingObjects) {
            const addressBuffer = lisk_cryptography_1.intToBuffer(unlock.delegateAddress.slice(0, -1), SIZE_UINT64);
            bufferArray.push(addressBuffer);
            const amountBuffer = lisk_cryptography_1.intToBuffer(unlock.amount.toString(), SIZE_INT64, 'big', true);
            bufferArray.push(amountBuffer);
            const unvoteHeightBuffer = lisk_cryptography_1.intToBuffer(unlock.unvoteHeight, SIZE_UINT32);
            bufferArray.push(unvoteHeightBuffer);
        }
        return Buffer.concat(bufferArray);
    }
    async prepare(store) {
        const addressArray = this.asset.unlockingObjects.map(unlock => ({
            address: unlock.delegateAddress,
        }));
        const filterArray = [
            {
                address: this.senderId,
            },
            ...addressArray,
        ];
        await store.account.cache(filterArray);
    }
    validateAsset() {
        const asset = this.assetToJSON();
        const schemaErrors = lisk_validator_1.validator.validate(unlockAssetFormatSchema, asset);
        const errors = errors_1.convertToAssetError(this.id, schemaErrors);
        for (const unlock of this.asset.unlockingObjects) {
            if (unlock.amount <= BigInt(0)) {
                errors.push(new errors_1.TransactionError('Amount cannot be less than or equal to zero', this.id, '.asset.votes.amount', unlock.amount.toString()));
            }
            if (unlock.amount % AMOUNT_MULTIPLIER_FOR_VOTES !== BigInt(0)) {
                errors.push(new errors_1.TransactionError('Amount should be multiple of 10 * 10^8', this.id, '.asset.votes.amount', unlock.amount.toString()));
            }
        }
        return errors;
    }
    async applyAsset(store) {
        const errors = [];
        for (const unlock of this.asset.unlockingObjects) {
            const sender = await store.account.get(this.senderId);
            const delegate = await store.account.getOrDefault(unlock.delegateAddress);
            if (!delegate.username) {
                errors.push(new errors_1.TransactionError('Voted account is not registered as delegate', this.id, '.asset.unlockingObjects.delegateAddress'));
                continue;
            }
            const waitingPeriod = getWaitingPeriod(sender, delegate, store.chain.lastBlockHeader.height, unlock);
            if (waitingPeriod > 0) {
                errors.push(new errors_1.TransactionError('Unlocking is not permitted as it is still within the waiting period', this.id, '.asset.unlockingObjects.unvoteHeight', waitingPeriod, 0));
            }
            const punishmentPeriod = utils_1.getPunishmentPeriod(sender, delegate, store.chain.lastBlockHeader.height);
            if (punishmentPeriod > 0) {
                errors.push(new errors_1.TransactionError('Unlocking is not permitted as delegate is currently being punished', this.id, '.asset.unlockingObjects.delegateAddress', punishmentPeriod, 0));
            }
            const unlockIndex = sender.unlocking.findIndex(obj => obj.amount === unlock.amount &&
                obj.delegateAddress === unlock.delegateAddress &&
                obj.unvoteHeight === unlock.unvoteHeight);
            if (unlockIndex < 0) {
                errors.push(new errors_1.TransactionError('Corresponding unlocking object not found', this.id, '.asset.unlockingObjects'));
                continue;
            }
            sender.unlocking.splice(unlockIndex, 1);
            sender.balance += unlock.amount;
            store.account.set(sender.address, sender);
        }
        return errors;
    }
    async undoAsset(store) {
        for (const unlock of this.asset.unlockingObjects) {
            const sender = await store.account.get(this.senderId);
            sender.balance -= unlock.amount;
            sender.unlocking.push(unlock);
            utils_1.sortUnlocking(sender.unlocking);
            store.account.set(sender.address, sender);
        }
        return [];
    }
}
exports.UnlockTransaction = UnlockTransaction;
UnlockTransaction.TYPE = 14;
//# sourceMappingURL=14_unlock_transaction.js.map